/* newplayer.c 

   Routines for handling players, sending messages to players, new players,
   etc.

 Chris Eleveld      (The Insane Hermit)
 Robert Slaven      (The Celestial Knight/Tigger)
 Jerry Bolton, Jr.  (Creel)
 Dave Mott          (Energizer Rabbit)

*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef POSIX
#  include <sys/select.h>
#endif
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include "lorien.h"

#define level(p,w)  ((SHOW(p) || w->seclevel >= p->seclevel) ? p->seclevel : 1)

#ifndef errno
extern int errno;
#endif

extern int getsock(), infromsock(), outtosock(), acceptcon(), atoi();

extern char *strcontains();

struct splayer *players;	/* the linked list of players records */
chan           *channels;
char            recvbuf[BUFSIZE];
char            sendbuf[OBUFSIZE];
int             numconnect;		/* number of people connected */

int numconnected() 
{
    int             i;
    struct splayer *curr;

    curr = players->next;
    i = 0;
    while (curr) {
	curr = curr->next;
	i++;
    }
    return i;
}

int
initplayerstruct()
{
	/* get a place to put our players */

    players = (struct splayer *) malloc(sizeof(struct splayer));

    if (players == (struct splayer *) 0) {
	fprintf(stderr, "lorien: Unable to allocate memory for players!\n");
	exit(1);
    }
    
    if ((channels = (chan *) malloc(sizeof(chan))) == (chan *) 0) {
	fprintf(stderr, "lorien: Unable to allocate memory for channels!\n");
	exit(1);
    }

    channels->next = (chan *) 0;
    strcpy(channels->name,DEFCHAN);
    channels->secure = 0;
    channels->count = 0;
    channels->visited = 0;

    players->next = (struct splayer *)0;

}

#ifdef CHECK_CHANNELS
int checkchannels(curr)
    chan *curr;
{
    int rc;

    if (!curr)
	return 1;

    if (curr->visited) {
	log("checkchannels():  found circular link in channel list.\n");
	return 0;
    }

    curr->visited = 1;
    rc = checkchannels(curr->next); 
#  ifdef FIX_CHANNEL_LINKS
    if (!rc) {
	curr->next = (chan) * 0;
	rc = 1;
    }
#  endif
    curr->visited = 0;
    return rc;
}
#endif

chan *
findchannel(name)
    char *name;
{
    chan *curr;

#ifdef CHECK_CHANNELS
    if (!checkchannels(channels)) {
	log("findchannel(): channel list incorrect.");
	sleep(2);
	abort(1);
    }
#endif

    curr = channels;

    while (curr) {
	if (!strncmp(curr->name,name,MAX_CHAN))
	    return curr;

	curr = curr->next;
    }

    return (chan *) 0;
}

chan *
newchannel(name)
    char *name;
{
    chan *newchan;

    newchan = (chan *) malloc(sizeof(chan));

    if (!newchan) 
	return newchan;

    newchan->secure = 0;
    newchan->visited = 0;
    strncpy(newchan->name,name,MAX_CHAN);

    newchan->name[MAX_CHAN] = 0;
    newchan->next = (chan *) 0;
    newchan->count = 0;  /* we wouldn't be building this if someone hadn't
			    requested it. */

    if (channels->next)
	newchan->next = channels->next;
    channels->next = newchan;

#ifdef CHECK_CHANNELS
    if (!checkchannels(channels)) {
	log("newchannel():  channel list incorrect.");
	sleep(2);
	abort(1);
    }
#endif

    return newchan;
}

void 
remove_channel(channel)
    chan *channel;
{
    chan *tmp;

#ifdef CHECK_CHANNELS
    if (!checkchannels(channels)) {
	log("remove_channel():  channel list incorrect on entry.");
	sleep(2);
	abort(1);
    }
#endif

    tmp = channels;

    if (tmp == channel)
	return; /* don't remove the main. */

    while (tmp->next) {
	if (tmp->next == channel) {
	    tmp->next = channel->next;
	    free(channel);
	    break;
	}
	tmp = tmp->next;
    }

#ifdef CHECK_CHANNELS
    if (!checkchannels(channels)) {
	log("remove_channel():  channel list incorrect on exit.");
	sleep(2);
	abort(1);
    }
#endif

}

int 
blocking(who)
    struct splayer *who;
{
    return (ban_findsite(who->onfrom) ? 1 : 0);
}

int
newplayer(s)
    int s;
{
    int             sock;
    int             i;
    struct splayer *buf;
    time_t          tmptime;
   
	/* find the tail of the linked list of players */

    buf = players;
    while (buf->next != (struct splayer *)0)
	buf = buf->next;
       
	/* create a new player struct(record) */
    buf->next = (struct splayer *) malloc(sizeof (struct splayer));
   
    if (buf->next == (struct splayer *) 0)  {
   
	    /*
	     * use the dummy record at the beginning of the list to accept and
	     * dump the incoming call because we're out of memory.
	     */

	sock = acceptcon(s, players->onfrom, &players->port);
	if (sock < 0)
	    return -1;
   
	sprintf(sendbuf,">> Unable to allocate memory for player record.\r\n");
	(void) outtosock(sock, sendbuf);
	(void) close(sock);
   
	    /* not fatal should let others continue to chat */
	return -1;
    }
   
    buf->next->next = (struct splayer *) 0; 

    sock = acceptcon(s, players->onfrom, &players->port);
    if (sock < 0)
	return (-1);
   
	/* Trust me it's 3 here and 4 there --- Jerry */

    if ((numconnected() + 1) >= (MAXCONN - 3)) {
	sprintf(sendbuf, ">> All %d connections are full.\r\n", MAXCONN -4);
	outtosock(sock,sendbuf);
	(void) close(sock);
	free(buf->next);
	buf->next = (struct splayer *) 0;
	return -1;
    }
   
    if (blocking(players)) {
	sprintf(sendbuf, ">> I'm sorry.  Someone at your site has angered the administrator.\r\nYour site is blocked.\r\nPlease try again later when the administrator isn't so angry.\r\n");

	outtosock(sock,sendbuf);
	(void) close(sock);
	free(buf->next);
	buf->next = (struct splayer *) 0;
	return -1;
    }
   
   
    tmptime = time((time_t *) 0);
    buf = buf->next;
    strcpy(buf->onfrom,players->onfrom);
    CLR_LEAVING(buf);
    CLR_VRFY(buf);
    SET_SHOW(buf);
    CLR_BEEPS(buf);
    CLR_HUSH(buf);
    CLR_ECHO(buf);
    CLR_WRAP(buf);
    CLR_SCREAM(buf);
    SET_INFO(buf);
    SET_MSG(buf);
    buf->port = players->port;
    buf->seclevel = BABYCO;
    buf->hilite = 0;
    buf->chnl = channels;
    buf->s = sock;
    strncpy(buf->name,DEFAULT_NAME, MAX_NAME);
    buf->name[MAX_NAME-1] = (char)0;
    buf->idle = tmptime;
    buf->cameon = tmptime;
    buf->next = (struct splayer *)0;
    buf->privs = CANDEFAULT;

    FD_ZERO(&buf->gags);
    memset(buf->pbuf,0,sizeof(buf->pbuf));

#ifndef NO_LOG_CONNECT
    sprintf(sendbuf,"Someone came on from %s on line %d\n",buf->onfrom,buf->s);
   
    log(sendbuf);
#endif
   
       
	/* welcome our new arrival */
   
    numconnect++;
    welcomeplayer(buf);
    channels->count++;
   
	/* let everyone else know who's here. */
    sprintf(sendbuf,">> New arrival on line %d from %s.\r\n", buf->s, buf->onfrom);
    sendall(sendbuf, ARRIVAL,0);
   
    return buf->s;
}

int
resetgags(line)
    int line;
{
    struct splayer *buf;
     
    buf = players;
    while (buf->next != (struct splayer *)0) {
	buf = buf -> next;
	  
	FD_CLR(line, &buf->gags);
    };
}

int
sendall(message, channel, line)
    char *message;
    chan *channel;
    int   line;
{
    struct splayer *buf,*before;
   
    buf = players;
    while (buf->next != (struct splayer *)0) {
           
	before = buf;
	buf = buf->next;

	if (channel == ALL) 
	    sendtoplayer(buf, message);
	else if (channel == INFORMATIONAL) {
	    if (INFO(buf))
		sendtoplayer(buf, message);
	}
	else if (channel == ARRIVAL) {
	    if (MSG(buf))
		sendtoplayer(buf,message);
	}
	else if (channel == DEPARTURE) {
	    if (buf->s == line) 
		before->next = buf->next;
	    else {

		    /* Un-gag the leaving player for the player we are processing. */

		FD_CLR(line, &buf->gags);

		if (MSG(buf))
		    sendtoplayer(buf,message);
	    }
	}
	else if (buf->chnl == channel) 
	    sendtoplayer(buf, message);
    }
}

int
setfds(needread)
    fd_set *needread;
{
    struct splayer *pplayer;
    int             max = 0;

    pplayer = players;
    while (pplayer->next != (struct splayer *)0) {
	pplayer = pplayer->next;
	if (pplayer->s > max)
	    max = pplayer->s;
	FD_SET(pplayer->s, needread);
    }
    return max;
}

int
processinput(pplayer)
    struct splayer *pplayer;
{
    time_t tmptime;
    char   *myptr = pplayer->pbuf;
    char   *line;

    tmptime = time((time_t *) 0);
    pplayer->idle = tmptime;

    while (line = strtok(myptr,"\r\n")) {

	myptr = (char *) 0;

	if (*line == '.' || *line == ',' || *line == '/')
	    handlecommand(pplayer,line);
	else {
	    if (SCREAM(pplayer)) { /* handle the ever-popular but annoying... */
		if (*line != '>') { /* > will redirect to channel. */
		    yell(pplayer,line); 
		    continue;
		} 
		    /* redirect to channel */
		line++; /* skip the '>' */
	    }
	    if (*line == ':' || *line == ';') 
		pose_it(pplayer,line);
	    else {
		sprintf(sendbuf,"(%d, %s) %s\r\n", pplayer->s, pplayer->name, 
			line);
		sendall(sendbuf,pplayer->chnl,0);
	    }
	}
    }
    memset(pplayer->pbuf,0,sizeof(pplayer->pbuf));
}

handleinput(needread)
    fd_set needread;
{
    char           *iptr;
    char           *bufptr;
    struct splayer *pplayer;
    struct splayer *tplayer;

    tplayer = players->next;

    while (tplayer != (struct splayer *)0) {

	pplayer = tplayer;
	tplayer = pplayer->next;

        if (LEAVING(pplayer)) { /* avoid the need to wait for input */
	    removeplayer(pplayer);
	    continue;
	}

	if (FD_ISSET(pplayer->s, &needread)) {

		/* check to see if they disconnected */

	    if (recvfromplayer(pplayer, recvbuf, BUFSIZE) == -1) {
		removeplayer(pplayer);
		continue;
	    }

	    recvbuf[BUFSIZE-1] = 0; /* ensure end of buffer. */
   
		/* now we have some data. */

	    iptr = recvbuf;
	    bufptr = strchr(pplayer->pbuf,(char) 0); 

	    while (*iptr) {
            
		if (strlen(pplayer->pbuf) >= BUFSIZE) {
		    sendtoplayer(pplayer, ">> fatal:  %d byte buffer overrun.\r\n");
		    recvbuf[0] = 0;
               
		    bufptr = pplayer->pbuf;
		    removeplayer(pplayer);
		    break;
		}

		if (*iptr == '\r' || *iptr == '\n') {

		    while (*iptr == '\r' || *iptr == '\n')
			iptr++;

		    if (!LEAVING(pplayer)) {
			processinput(pplayer);
		    }

			/* now we check again after processing the input. */

		    if (LEAVING(pplayer)) {
			removeplayer(pplayer);
			break;
		    }

		    bufptr = pplayer->pbuf;
		}
		else {

			/* It's time to get rid of the caps, if necessary. */

		    if (isascii(*iptr) && isprint(*iptr)) {

			if (!(pplayer->privs & CANCAPS)) 
			    *bufptr = tolower(*iptr);
			else
			    *bufptr = *iptr;
			bufptr++;
		    }
		    iptr++;

		}
	    }
	}
    }
}

int
removeplayer(player)
    struct splayer *player;	/* a pointer to the player to be nuked */
{

    player->chnl->count--;

    if (player->chnl->count <= 0)
	remove_channel(player->chnl);

    sprintf(sendbuf, ">> line %d(%s) just left.\r\n", player->s, player->name);

    if ((player->s <= (MAXCONN - 3))) {
	SET_LEAVING(player);
	sendall(sendbuf, DEPARTURE, player->s);
    }

    shutdown(player->s, 2);

    (void) close(player->s);
    free((struct splayer  *) player);
    numconnect--;
}

char *
idlet(idle)
    long idle;
{
    static char buf[11];
    long timenoc;
    time_t tmptime;
    long hrs,mins,secs;
    
    tmptime = time((time_t *) 0);
    timenoc = tmptime - idle;
    secs = timenoc % 60;
    mins = (timenoc / 60) % 60;
    hrs = timenoc / 3600;

    if (hrs) {
	if (mins)
	    sprintf(buf, "%dh%dm", hrs, mins);
	else
	    sprintf(buf, "%dh", hrs);
    } else if (mins) {
	if (secs)
	    sprintf(buf, "%dm%ds", mins, secs);
	else
	    sprintf(buf, "%dm", mins);
    } else if (secs) {
	sprintf(buf, "%ds", secs);
    } else 
	*buf = '\000';

    return buf;
}

/* Creel 6 Sep 94 */
int 
wholist(pplayer, target)
    struct splayer *pplayer;
    char *target;
{
    struct splayer *buf;
    int count = 0;
    int match;

    buf = players->next;

#ifdef ONFROM_ANY
    sprintf(sendbuf,"%-4s %-27s %-13s %-6s %-25s\r\n","Line","Name","Channel","Idle","Doing  ");
#else
    sprintf(sendbuf,"%-4s %-27s %-13s %-6s %-25s\r\n","Line","Name","Channel","Idle","On From");
#endif
    sendtoplayer(pplayer, sendbuf);
    sendtoplayer(pplayer, LINE);

    while (buf) {
	match = 0;
	if (target) {
	    sprintf(sendbuf,"%d",buf->s); 
	    if (atoi(target) && !strchr(target,'.')) {
		if (atoi(target) == buf->s)
		    match = 1; 
	    }
	    else if (strcontains(buf->name,target,SUB) ||
		     strcontains(buf->chnl->name,target,SUB) ||
		     strcontains(buf->onfrom, target,SUB) ||
		     strcontains(sendbuf,target,SUB))
		match = 1;
	}

	sprintf(sendbuf,"%c%c%-2d %-27.27s %-13.13s %-6s %-25.25s\r\n",
		HUSH(buf) ? 'H' : ' ',
		(buf->chnl->secure) ? 'S': ' ',
		buf->s,
		buf->name,
		buf->chnl->name,
		idlet(buf->idle),
		buf->onfrom);

	if ((!target) ||
	    (target && match)) {

	    sendtoplayer(pplayer, sendbuf);
	    count++;
	};

	buf = buf->next;
    }

    sendtoplayer(pplayer,LINE);

    if (count == 1)
	sprintf(sendbuf,">> %d record displayed.\r\n",count);   
    else
	sprintf(sendbuf,">> %d records displayed.\r\n",count);   

    sendtoplayer(pplayer, sendbuf);

}

int 
wholist2(pplayer,target)
    struct splayer *pplayer;
    char *target;
{
    struct splayer *buf;
    int count = 0;
    char vrfy[4][4];
    int match;
    
    strcpy(vrfy[0],"No");
    strcpy(vrfy[1],"Yes");

    buf = players->next;

    sprintf(sendbuf,"%-4s %-30s %-8s %-8s %-8s %-8s\r\n","Line","Name","On For","Verified","Port","Level");
    sendtoplayer(pplayer, sendbuf);
    sendtoplayer(pplayer,LINE);

    while (buf) {
	match = 0;
	if (target) {
	    sprintf(sendbuf,"%d",buf->s); 
	    if (atoi(target) && !strchr(target,'.')) {
		if (atoi(target) == buf->s)
		    match = 1; 
	    }
	    else if (strcontains(buf->name,target,SUB) ||
		     strcontains(sendbuf,target,SUB) ||
		     strcontains(buf->onfrom, target,SUB))
		match = 1;
	}

	sprintf(sendbuf,"%c%c%-2d %-30.30s %-8s %-8s %-8d ",
		HUSH(buf) ? 'H' : ' ',
		(buf->chnl->secure) ? 'S' : ' ',
		buf->s,
		buf->name,
		timelet(buf->cameon,2),
		vrfy[VRFY(buf) ? 1 : 0],
		buf->port,
		level(buf,pplayer));

	if ((!target) ||
	    (target && match)) {

	    sendtoplayer(pplayer, sendbuf);

	    sprintf(sendbuf,"%-8d\r\n",(pplayer->seclevel > 0) ? level(buf,pplayer) : 1);
	    sendtoplayer(pplayer,sendbuf);

	    count++;
	};

	buf = buf->next;
    }

    sendtoplayer(pplayer,LINE);

    if (count == 1)
	sprintf(sendbuf,">> %d record displayed.\r\n",count);   
    else
	sprintf(sendbuf,">> %d records displayed.\r\n",count);   

    sendtoplayer(pplayer, sendbuf);

}

int 
wholist3(pplayer)
    struct splayer *pplayer;
{
    struct splayer *buf;
    int count = 0;
    
    buf = players->next;

    sendtoplayer(pplayer, LINE);

    while (buf) {
	sprintf(sendbuf,"%2d%c)%-14.14s",buf->s, 
		HUSH(buf) ? 'H' : ' ',
		buf->name);
	sendtoplayer(pplayer, sendbuf);
	count++;
	if (!(count % 4))
	    sendtoplayer(pplayer,"\r\n");

	buf = buf->next;
    }

    if (count % 4)
	sendtoplayer(pplayer,"\r\n");

    sendtoplayer(pplayer,LINE);

    if (count == 1)
	sprintf(sendbuf,">> %d record displayed.\r\n",count);   
    else
	sprintf(sendbuf,">> %d records displayed.\r\n",count);   

    sendtoplayer(pplayer, sendbuf);
   
}

int
cleanupbuf()
{
    char tmpbuf[BUFSIZE];
    char *newc, *old;

    old = recvbuf;
    newc = tmpbuf;
    while (*old != '\000') {
	if ((!iscntrl(*old)) || (*old == '\n') || (*old == '\r')) {
	    *newc = *old;
	    newc++;
	}
	old++;
    }
    *newc = '\000';
    strcpy(recvbuf, tmpbuf);
}

int
recvfromplayer(who, data, howmuch)
    struct splayer *who;
    char *data;
    int howmuch;
{
    int error;

    error = infromsock(who->s, data, howmuch);
	/* remove nasty chars */
    (void) cleanupbuf();

    return error;
}


int 
isgagged(who, line)
    struct splayer *who;
    int line;
{
    if ((line < 1) || (line >= MAXCONN))
	return 0; 
    else return( FD_ISSET(line,&who->gags));
}

char *
wrap(s)
    char *s;
{
    static char wrapped[BUFSIZE*2];
    int i = 0;  /* the number of characters copied. */
    int j = 0;  /* the index into wrapped buffer */
    int k = 0;

    memset(wrapped,0,BUFSIZE * 2);
    while (*s) {
	wrapped[j] = *s;
	j++;
	i++;
	if (!((i + k + 1) % 80)) {
	    strcat(wrapped,"\r\n   ");
	    j += 5;
	    k += 3;
	}
	s++;
    }
      
    return wrapped;  
      
}

int 
sendtoplayer(who, message)
    struct splayer *who;
    char *message;
{
    int line;
    int end = 1;
    int gagck = 0; 

    if (LEAVING(who))
	return 0;
/* If it is a message from a player, then we want to 
   listen to the returned value from isgagged, otherwise, go ahead
   and do isgagged because it takes all of 2 nanoseconds, but ignore the 
   result. */

    if (message[0] == '(') {
        while (!isdigit(message[end]) && message[end])
	    end++;
	line = atoi(&message[end]);
	gagck = 1;
    };
            
/* if we are not checking for gags, or we are and the sender is not gagged.
 */

    if (! (gagck && isgagged(who,line)) ) 
         
        if (outtosock(who->s, WRAP(who) ? wrap(message) : message) == -1) { /* no one there */
	    SET_LEAVING(who);
	    return 0;
        };
    return 1;
}

int
welcomeplayer(pplayer)
    struct splayer *pplayer;
{
    FILE *fpWELCOME;
    char  ibuf[BUFSIZE];

    if ((fpWELCOME = fopen(WELCOMEFILE, "r")) == (FILE *)0) {
	(void) sendtoplayer(pplayer, "Unable to open welcome file.\r\n");
	fprintf(stderr, "Unable to open welcome file %s.\n", WELCOMEFILE);
	return 0;
    }

    while (fgets(ibuf, BUFSIZE, fpWELCOME) != NULL) {
        sprintf(sendbuf,"%s\r",ibuf);
	(void) sendtoplayer(pplayer, sendbuf);
    }
    
    (void) fclose(fpWELCOME);

    sendtoplayer(pplayer, "This site is running Lorien ");
    sendtoplayer(pplayer, VERSION);
    sendtoplayer(pplayer, "\r\n");
    return 1;
}

struct splayer
*lookup(linenum)
    int linenum;
{
    struct splayer *tmp=players;

    if (linenum == 0)
	return (struct splayer *)0;
    
    while (tmp->next != (struct splayer *)0) {
	tmp = tmp->next;
	if (tmp->s == linenum)
	    return tmp;
    }

    return (struct splayer *)0;
}

int
setname(linenum, name)
    int linenum;
    char *name;
{
    char *buf=name;
    struct splayer *tmp;

	/* find out who to set */
    tmp = lookup(linenum);
    if (tmp == (struct splayer *)0) {
	return -1;
    } else {
	    /* check for returns and newlines */
	if ((buf = (char *) index(name, '\r')) != (char *)0) {
	    *buf = '\000';
	}
	if ((buf = (char *) index(name, '\n')) != (char *)0) {
	    *buf = '\000';
	}

	buf= name;
        buf = (char *) skipspace(buf);
	
	if (*buf != '\000') {
	    (void) strncpy(tmp->name, buf, MAX_NAME-1);
            tmp->name[MAX_NAME - 1] = 0;
        }
	else {
	    sprintf(sendbuf, ">> As if! Pick a real name!\r\n");
	    sendtoplayer(tmp, sendbuf);
	}
    }
    return 0;
}

